/*
 * Copyright (c) 2015-2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Tue Jun  2 2015
 */
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#if defined(_WIN32) || defined(_WIN64)
# include <winsock2.h>
# include <ws2tcpip.h>
#else /* posix */
# include <sys/socket.h>

# include <netdb.h>

# include <netinet/tcp.h>
#endif

#include "libgenomix.h"


/* --- local data ---------------------------------------------------------- */

static struct connections {
  const char *host;
  const char *session;
  socket_s cmdfd, evtfd;

  struct http_rqst rqst;
  struct http_event evt;
} *connections;

static uint8_t nconnection;

static socket_s	genomix_socket(const char *host, const char *port, char *emsg,
			size_t emsgsize);
static int	genomix_http_buffer(struct http_buffer *buffer);


/* --- genomix_connect ----------------------------------------------------- */

int
genomix_connect(char *hostport, char *reqsession, char *emsg, size_t emsgsize)
{
  struct connections conn;
  char *host, *port;
  char uri[256];
  int M;
  float v;
  int s;

#if defined(_WIN32) || defined(_WIN64)
  { /* setup sockets */

    WSADATA supported;
    s = WSAStartup(MAKEWORD(2, 2), &supported);
    if (s) {
      snprintf(emsg, emsgsize, "could not initialize sockets");
      return -1;
    }
    if (LOBYTE(supported.wVersion) != 2 || HIBYTE(supported.wVersion) != 2) {
      snprintf(emsg, emsgsize,
               "could not find a useable version of winsock.dll");
      WSACleanup();
      return -1;
    }
  }
#endif

  if (hostport) {
    host = hostport;
    if ((port = strchr(host, ':'))) {
      *port++ = '\0';
    } else
      port = "8080";
  } else {
    host = "localhost";
    port = "8080";
  }
  conn.host = strdup(host);
  conn.session = NULL;
  conn.cmdfd = conn.evtfd = INVALID_SOCKET;
  if (genomix_http_buffer(&conn.rqst.uri)) goto err;
  if (genomix_http_buffer(&conn.rqst.buffer)) goto err;
  if (genomix_http_buffer(&conn.rqst.args)) goto err;
  if (genomix_http_buffer(&conn.evt.buffer)) goto err;

  /* connect to genomix */
  conn.cmdfd = genomix_socket(host, port, emsg, emsgsize);
  if (conn.cmdfd == INVALID_SOCKET) goto err;

  s = http_transaction(conn.cmdfd, &conn.rqst, "get", "/version", NULL);
  if (s < 200 || s >= 300) {
    snprintf(emsg, emsgsize, "unable to get genomixd server version: %s",
             conn.rqst.body);
    goto err;
  }
  if (sscanf(conn.rqst.body, "%d.%f", &M, &v) != 2) {
    snprintf(emsg, emsgsize, "not a genomixd server");
    goto err;
  }
  if (M < 1 || (M == 1 && v < 6.09)) {
    snprintf(emsg, emsgsize,
             "unsupported genomixd server version %d.%g, need 1.6.1", M, v);
    goto err;
  }


  /* request a session */
  snprintf(uri, sizeof(uri), "/session/start?s=%s", reqsession?reqsession:"");
  s = http_transaction(conn.cmdfd, &conn.rqst, "get", uri, NULL);
  if (s < 200 || s >= 300) {
    snprintf(emsg, emsgsize, "%s", conn.rqst.body);
    goto err;
  }
  conn.session = strdup(conn.rqst.body);


  /* event socket */
  conn.evtfd = genomix_socket(host, port, emsg, emsgsize);
  if (conn.evtfd == INVALID_SOCKET) return -1;

  snprintf(uri, sizeof(uri), "/session/listen?s=%s", conn.session);
  s = http_transaction(conn.evtfd, &conn.rqst, "get", uri, NULL);
  if (s < 200 || s >= 300) {
    snprintf(emsg, emsgsize, "%s", conn.rqst.body);
    goto err;
  }


  /* register connection */
  for (s = 0; s < nconnection; s++)
    if (connections[s].cmdfd == INVALID_SOCKET) break;
  if (s >= nconnection) {
    struct connections *c = realloc(connections, (s+1)*sizeof(*c));
    if (!c) goto err;
    connections = c;
    nconnection = s+1;
  }
  connections[s] = conn;

  return s;
err:
  if (conn.host) free((void *)conn.host);
  if (conn.session) free((void *)conn.session);
  if (conn.rqst.uri.data) free(conn.rqst.uri.data);
  if (conn.rqst.buffer.data) free(conn.rqst.buffer.data);
  if (conn.rqst.args.data) free(conn.rqst.args.data);
  if (conn.evt.buffer.data) free(conn.evt.buffer.data);

#if defined(_WIN32) || defined(_WIN64)
  if (conn.cmdfd != INVALID_SOCKET) closesocket(conn.cmdfd);
  if (conn.evtfd != INVALID_SOCKET) closesocket(conn.evtfd);
  WSACleanup();
#else
  if (conn.cmdfd != INVALID_SOCKET) close(conn.cmdfd);
  if (conn.evtfd != INVALID_SOCKET) close(conn.evtfd);
#endif

  return -1;
}


/* --- genomix_disconnect -------------------------------------------------- */

int
genomix_disconnect(int id)
{
  struct connections *conn;
  int i, remain;

  for (i=remain = 0; i<nconnection; i++)
    if (connections[i].cmdfd != INVALID_SOCKET) remain++;

  if (id >= nconnection) return remain;
  conn = connections + id;

  if (conn->host) free((void *)conn->host);
  conn->host = NULL;
  if (conn->session) free((void *)conn->session);
  conn->session = NULL;
  if (conn->rqst.uri.data) free(conn->rqst.uri.data);
  conn->rqst.uri.data = NULL;
  if (conn->rqst.buffer.data) free(conn->rqst.buffer.data);
  conn->rqst.buffer.data = NULL;
  if (conn->rqst.args.data) free(conn->rqst.args.data);
  conn->rqst.args.data = NULL;
  if (conn->evt.buffer.data) free(conn->evt.buffer.data);
  conn->evt.buffer.data = NULL;

#if defined(_WIN32) || defined(_WIN64)
  if (conn->evtfd != INVALID_SOCKET) closesocket(conn->evtfd);
  if (conn->cmdfd != INVALID_SOCKET) closesocket(conn->cmdfd);
  WSACleanup();
#else
  if (conn->evtfd >= 0) close(conn->evtfd);
  if (conn->cmdfd >= 0) close(conn->cmdfd);
#endif
  conn->evtfd = INVALID_SOCKET;
  conn->cmdfd = INVALID_SOCKET;

  return remain - 1;
}


/* --- genomix_get --------------------------------------------------------- */

const char *
genomix_get(int id, const char *cmd, const char *args[],
            char *emsg, size_t emsgsize)
{
  struct connections *conn;

  conn = connections + id;
  if (id >= nconnection || conn->cmdfd == INVALID_SOCKET) {
    snprintf(emsg, emsgsize, "no such connection %d", id);
    return NULL;
  }

  if (http_get(conn->cmdfd, &conn->rqst, conn->session, cmd, args)) {
    snprintf(emsg, emsgsize, "%s", conn->rqst.body);
    return NULL;
  }

  return conn->rqst.body;
}


/* --- genomix_doevent ----------------------------------------------------- */

void
genomix_doevent(int timeout)
{
  struct timeval tout;
  fd_set fds;
  int maxfd;
  int i, s;

  FD_ZERO(&fds);
  maxfd = 0;
  for (i=0; i<nconnection; i++) {
    if (connections[i].evtfd == INVALID_SOCKET) continue;

    FD_SET(connections[i].evtfd, &fds);
    if (connections[i].evtfd > maxfd) maxfd = connections[i].evtfd;
  }

  if (timeout >= 0) {
    tout.tv_sec = timeout/1000.;
    tout.tv_usec = 1000 * (timeout - tout.tv_sec * 1000);
    s = select(maxfd+1, &fds, NULL, NULL, &tout);
  } else
    s = select(maxfd+1, &fds, NULL, NULL, NULL);
  if (s == SOCKET_ERROR || s == 0) return;

  for (i=0; i<nconnection; i++) {
    if (connections[i].evtfd == INVALID_SOCKET) continue;

    if (FD_ISSET(connections[i].evtfd, &fds)) {
      struct connections *c = &connections[i];
      do {
        if (http_event(c->evtfd, &c->evt)) break;
        if (c->evt.event != -1) {
          if (!strcmp(c->evt.buffer.data + c->evt.event, "request"))
            genomix_event(c->evt.buffer.data + c->evt.data);
          else if (!strcmp(c->evt.buffer.data + c->evt.event, "quit")) {
            genomix_disconnect(i);
            break;
          }
        }
      } while(c->evt.buffer.end > c->evt.buffer.start);
    }
  }
}


/* --- genomix_socket ------------------------------------------------------ */

static socket_s
genomix_socket(const char *host, const char *port, char *emsg, size_t emsgsize)
{
  struct addrinfo hints, *res, *res0;
  const char *cause;
  socket_s fd;
  int e, ecause;

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  e = getaddrinfo(host, port, &hints, &res0);
  if (e) {
    snprintf(emsg, emsgsize, "%s", gai_strerror(e));
    return INVALID_SOCKET;
  }

  fd = INVALID_SOCKET;
  for (res = res0; res; res = res->ai_next) {
    if (res->ai_family != AF_INET && res->ai_family != AF_INET6) continue;

    fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (fd == INVALID_SOCKET) {
      cause = "socket";
#if defined(_WIN32) || defined(_WIN64)
      ecause = WSAGetLastError();
#else
      ecause = errno;
#endif
      continue;
    }

    if (connect(fd, res->ai_addr, res->ai_addrlen) < 0) {
      cause = "connect";
#if defined(_WIN32) || defined(_WIN64)
      ecause = WSAGetLastError();
      closesocket(fd);
#else
      ecause = errno;
      close(fd);
#endif
      fd = INVALID_SOCKET;
      continue;
    }

    setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (char *)&(int){1}, sizeof(int));
    break;
  }
  freeaddrinfo(res0);

  if (fd == INVALID_SOCKET) {
#if defined(_WIN32) || defined(_WIN64)
    char s[128];
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, ecause, 0,
                  s, sizeof(s), NULL);
    snprintf(emsg, emsgsize, "%s: %s", cause, s);
#else
    snprintf(emsg, emsgsize, "%s: %s", cause, strerror(ecause));
#endif
    return INVALID_SOCKET;
  }

/* set non-blocking */
#if defined(_WIN32) || defined(_WIN64)
  if (ioctlsocket(fd, FIONBIO, &(unsigned long){1}) != NO_ERROR)
    /* ignore error */;
#else
  {
    int f = fcntl(fd, F_GETFL);
    if (f != -1) {
      f |= O_NONBLOCK;
      fcntl(fd, F_SETFL, f);
    }
  }
#endif

  return fd;
}


/* --- genomix_http_buffer ------------------------------------------------- */

static int
genomix_http_buffer(struct http_buffer *buffer)
{
  buffer->data = malloc(64);
  if (!buffer->data) return -1;

  buffer->len = 64;
  buffer->end = buffer->data;
  buffer->start = buffer->data;
  return 0;
}
